/***************************************************************************
 *   Copyright (C) 2008 by David Stockwell   *
 *   dstockwell@frequency-one.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <iostream>
#include <cstdlib>
#include <stdlib.h>

#include "gdbus.h"

#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <glib-object.h>

using namespace std;

static	GMainLoop	*mainLoop=NULL;

static	DBusMessage *agent_authorize(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	char *objectPath = (char *) malloc(128);
	char *objectUUID = (char *) malloc(128);

	if (dbus_message_get_args(msg, NULL,
			DBUS_TYPE_OBJECT_PATH, &objectPath,
			DBUS_TYPE_STRING, &objectUUID,
			DBUS_TYPE_INVALID) == FALSE) {
//		free(objectPath);
//		free(objectUUID);
		return NULL;
	}
	cout << "Authorize received..." << objectPath << "-" << objectUUID << endl;

//	free(objectPath);
//	free(objectUUID);

	DBusMessage *reply = dbus_message_new_method_return(msg);
	if (!reply) return NULL;

	return reply;
}

static	DBusMessage *agent_cancel(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	cout << "Cancel received..." << endl;

	DBusMessage *reply = dbus_message_new_method_return(msg);
	if (!reply) return NULL;

	return reply;
}

static	DBusMessage *agent_confirm_mode_change(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	char *mode = (char *) malloc(128);

	if (dbus_message_get_args(msg, NULL,
			DBUS_TYPE_STRING, &mode,
			DBUS_TYPE_INVALID) == FALSE) {
//		free(mode);
		return NULL;
	}
	cout << "ConfirmModeChange received..." << mode << endl;

//	free(mode);
	DBusMessage *reply = dbus_message_new_method_return(msg);
	if (!reply) return NULL;

	return reply;
}

static	DBusMessage *agent_display_passkey(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	char *objectPath = (char *) malloc(128);
	uint passKey;
	short entered;

	if (dbus_message_get_args(msg, NULL,
			DBUS_TYPE_OBJECT_PATH, &objectPath,
			DBUS_TYPE_UINT32, &passKey,
			DBUS_TYPE_BYTE, &entered,
			DBUS_TYPE_INVALID) == FALSE) {
//		free(objectPath);
		return NULL;
	}
	cout << "DisplayPasskey received..." << objectPath << "-" <<  passKey << "-" << entered << endl;

//	free(objectPath);

	DBusMessage *reply = dbus_message_new_method_return(msg);
	if (!reply) return NULL;

	return reply;
}

static	DBusMessage *agent_release(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	cout << "Release received...exiting" << endl;

	DBusMessage *reply = dbus_message_new_method_return(msg);
	if (!reply) return NULL;

	g_main_loop_quit(mainLoop);
	return reply;
}

static	DBusMessage *agent_request_confirmation(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	char *objectPath = (char *) malloc(128);
	uint passKey;

	if (dbus_message_get_args(msg, NULL,
			DBUS_TYPE_OBJECT_PATH, &objectPath,
			DBUS_TYPE_UINT32, &passKey,
			DBUS_TYPE_INVALID) == FALSE) {
//		free(objectPath);
		return NULL;
	}
	cout << "RequestConfirmation received..." << objectPath << "-" <<  passKey << endl;

//	free(objectPath);

	DBusMessage *reply = dbus_message_new_method_return(msg);
	if (!reply) return NULL;

	return reply;
}

static	DBusMessage *agent_request_passkey(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	char *objectPath = (char *) malloc(128);
	uint passKey = 0000;			// For Testing Only

	if (dbus_message_get_args(msg, NULL,
			DBUS_TYPE_OBJECT_PATH, &objectPath,
			DBUS_TYPE_INVALID) == FALSE) {
//		free(objectPath);
		return NULL;
	}
	cout << "RequestPasskey received..." << objectPath << endl;

//	free(objectPath);

	DBusMessage *reply = dbus_message_new_method_return(msg);
	if (!reply) return NULL;

	if (!dbus_message_append_args(reply, DBUS_TYPE_UINT32, passKey, DBUS_TYPE_INVALID)) {
		dbus_message_unref(reply);
		return NULL;
	}
	return reply;
}

static	DBusMessage *agent_request_pincode(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	char *objectPath = (char *) malloc(128);
	const char * pinCode = "Bluez";		// For testing only...

	if (dbus_message_get_args(msg, NULL,
			DBUS_TYPE_OBJECT_PATH, &objectPath,
			DBUS_TYPE_INVALID) == FALSE) {
//		free(objectPath);
		return NULL;
	}
	cout << "RequestPinCode received..." << objectPath << endl;

//	free(objectPath);

	DBusMessage *reply = dbus_message_new_method_return(msg);
	if (!reply) return NULL;

	if (!dbus_message_append_args(reply, DBUS_TYPE_STRING, &pinCode, DBUS_TYPE_INVALID)) {
		dbus_message_unref(reply);
		return NULL;
	}
	return reply;
}

/* from doc/agent-api.txt BlueZ 4.0 API */

static	GDBusMethodTable agent_methods[] = {
	{ "Authorize",		"os",	"",	agent_authorize			},
	{ "Cancel",		"",	"",	agent_cancel			},
	{ "ConfirmModeChange",	"s",	"",	agent_confirm_mode_change	},
	{ "DisplayPasskey",	"ouy",	"",	agent_display_passkey		},
	{ "Release",		"",	"",	agent_release			},
	{ "RequestConfirmation","ou",	"",	agent_request_confirmation	},
	{ "RequestPasskey",	"o",	"u",	agent_request_passkey		},
	{ "RequestPinCode",	"o",	"s",	agent_request_pincode		},
	{ }
};

/* Note that there are no signals defined.  If an answer is not positive, an error is thrown. */

int	agent_install(DBusGConnection *dBusG, string agentObjectPath)
{
	DBusError	dError;

	DBusConnection	*dBus=dbus_g_connection_get_connection(dBusG);
	dbus_error_init(&dError);
/*
	With the connection to a the System Bus, register the object for the Agent, then register the interface.
*/
	if (g_dbus_register_object(dBus, agentObjectPath.c_str())) {
		cout << "AgentPath " << agentObjectPath << " REGISTERED" << endl;
	} else {
		cout << "Failed to register AgentPath: " << agentObjectPath;
		if (dbus_error_is_set(&dError)) {
			cout << dError.message;
			dbus_error_free(&dError);
		}
		cout << endl;
		return EXIT_FAILURE;
	}

	if (g_dbus_register_interface(dBus, agentObjectPath.c_str(), "org.bluez.Agent",
			agent_methods, NULL, NULL, NULL, NULL)) {
		cout << "AgentInterface for " << agentObjectPath << " REGISTERED" << endl;
	} else {
		cout << "Failed to register interface for Agent: " << agentObjectPath;
		if (dbus_error_is_set(&dError)) {
			cout << dError.message;
			dbus_error_free(&dError);
		}
		cout << endl;
		return EXIT_FAILURE;
	}

	return	EXIT_SUCCESS;
}

int	agent_uninstall(DBusGConnection *dBusG, string agentObjectPath)
{
	DBusConnection	*dBus=dbus_g_connection_get_connection(dBusG);

	g_dbus_unregister_interface(dBus, agentObjectPath.c_str(), "org.bluez.Agent");
	g_dbus_unregister_object(dBus, agentObjectPath.c_str());
	g_dbus_cleanup_connection(dBus);
	return	EXIT_SUCCESS;
}
